import torch


def clip_gradient(grad, k):
    """Clips the top k% of the gradients in terms of absolute value.

    Args:
        grad (torch.Tensor): The gradients to be clipped.
        k (float): The percentage of gradients to keep.

    Returns:
        torch.Tensor: The clipped gradients.
    """
    if len(grad.size()) < 2 or grad.size(1) < 2:
        return grad

    if k >= 1.0:
        return grad

    # Compute the threshold
    #print(f"grad: {grad}, k: {k}, grad.size(1): {grad.size(1)}, int(grad.size(1) * (1 - k))): {int(grad.size(1) * (1 - k))}")
    threshold = torch.kthvalue(torch.abs(grad), int(grad.size(1) * (1 - k)), dim=1)[0]

    # Clip the gradients
    clipped_grad = torch.clamp(torch.abs(grad), max=threshold.unsqueeze(1))
    clipped_grad *= torch.sign(grad)

    return clipped_grad


import torch

from clip import clip_gradient


class CoordinateDoWG(torch.optim.Optimizer):
    """Implements CDoWG-- a coordinate-wise version of DoWG.

        Args:
            params (iterable): iterable of parameters to optimize or dicts defining
                parameter groups
            eps (float, optional): term added to the denominator to improve
                numerical stability (default: 1e-8)
        """

    def __init__(self, params, epsilon=1e-8, clip=0.5, *args, **kwargs):
        defaults = dict(r_epsilon=epsilon, lr=1, clip=clip)
        super(CoordinateDoWG, self).__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            with torch.no_grad():
                for p in group['params']:
                    if p.grad is None:
                        continue
                    grad = p.grad.data
                    state = self.state[p]

                    # Initialize state variables
                    if 'x0' not in state:
                        state['x0'] = torch.clone(p).detach()
                    if 'rt2' not in state:
                        state['rt2'] = torch.zeros_like(p.data).add_(1e-8)
                    if 'vt' not in state:
                        state['vt'] = torch.zeros_like(p.data)

                    state['rt2'] = torch.max(state['rt2'], (p - state['x0']) ** 2)
                    rt2, vt = state['rt2'], state['vt']
                    vt.add_(rt2 * grad ** 2)
                    gt_hat = rt2 * clip_gradient(p, group['clip'])
                    denom = vt.sqrt().add_(group['epsilon'])
                    p.data.addcdiv_(gt_hat, denom, value=-1.0)
        return loss
    





import torch

from clip import clip_gradient


class ScalarDoWG(torch.optim.Optimizer):
    def __init__(self, params, epsilon=1e-4, clip=0.5, *args, **kwargs):
        defaults = dict(r_epsilon=epsilon, lr=1, clip=clip)
        self.epsilon = epsilon
        super(ScalarDoWG, self).__init__(params, defaults)

    def step(self):
        state = self.state

        with torch.no_grad():
            device = self.param_groups[0]["params"][0].device

            if "rt2" not in state:
                state["rt2"] = torch.Tensor([self.epsilon]).to(device)
            if "vt" not in state:
                state["vt"] = torch.Tensor([0]).to(device)

            grad_sq_norm = torch.Tensor([0]).to(device)
            curr_d2 = torch.Tensor([0]).to(device)

            for idx, group in enumerate(self.param_groups):
                group_state = state[str(idx)]  # convert idx to a string
                if "x0" not in group_state:
                    group_state["x0"] = [torch.clone(p) for p in group["params"]]

                grad_sq_norm += torch.stack(
                    [(p.grad**2).sum() for p in group["params"]]
                ).sum()
                curr_d2 += torch.stack(
                    [
                        ((p - p0) ** 2).sum()
                        for p, p0 in zip(group["params"], group_state["x0"])
                    ]
                ).sum()

            state["rt2"] = torch.max(state["rt2"], curr_d2)
            state["vt"] += state["rt2"] * grad_sq_norm
            rt2, vt = state["rt2"], state["vt"]

            for group in self.param_groups:
                for p in group["params"]:
                    gt_hat = rt2 * clip_gradient(p.grad.data, group["clip"])
                    denom = torch.sqrt(vt).add_(group["epsilon"])
                    p.data.addcdiv_(gt_hat, denom, value=-1.0)
        return None
